from defichainUtils.utils import ParamType
from defichainUtils.utils import utils,CustomTxDecode
from defichainUtils.utils import CustomTx
# from utils import ParamType
# from utils import utils,CustomTxDecode
# from utils import CustomTx
from dataclasses import dataclass


VERSION = 4
SEQUENCE = 'ffffffff'
DECIMALS = 100000000
def compose(txs:list):
    '''
    List of DfTx Class objects. Can be used for output (2.) in createrawtransaction (RPC DefiChain) 
    '''
    res = []
    for tx in txs:
        res.append(tx.hex())
    return res

def createRawTransaction(input:list,output:list,shareAddress:str,locktime:int=0,replaceable:bool=False):
    '''
    For Witness Transactions and BECH32 Addresses Only!!

    input transactions is list of dictionaries and need fields: txid, vout, amount

    output list are hex endcoded transactions. 
    UTXO output is calculated automatically and amount minus fee is sent to shareAddress.
    '''
    amount = 0
    res = ''
    res = res + utils.int2hex(VERSION,'little',4)
    res = res + utils.int2hex(len(input),'little')
    # (Unspent) UTXO
    for i in input:
        res = res + utils.convert_hex(i['txid'],'big','little')
        res = res + utils.int2hex(int(i['vout']),'little',4)
        res = res + '00'
        res = res + utils.generateSequence(SEQUENCE,locktime,replaceable)
        amount = amount + float(i['amount'])
    
    res = res + utils.int2hex(len(output)+1,'little')
    # OP_RETURN Transactions
    for i,o in enumerate(output):
        # '00' is seperator between transactions
        if i > 0:
            res = res + '00'
        res = res + utils.int2hex(0,'little',8)
        res = res + utils.int2hex(int(len(o)/2),'little')
        res = res + o

    # UTXO Output
    # '00' is seperator between transactions
    rbf = 0
    if replaceable:
        rbf = 1
    
    # TODO: signrawtransaction and get number of bytes to calculate vSize.
    dummy = '00' + utils.int2hex(0,'little',8) + utils.encodeBech32AddressToHex(shareAddress) + utils.int2hex(rbf,'little',1) + utils.int2hex(locktime,'little',4)
    
    size = int(len(res+dummy)/2)
    #size = int(736/2)
    size = int(580/2)
    # Witness data is 108 Bytes; vsize is "size" plus "3 time size minus witness data" divided by four. https://en.bitcoinwiki.org/wiki/Block_weight#Detailed_example ; https://bitcointalk.org/index.php?topic=5276203.0
    vSize = size - (108 * 3 / 4)
    print(utils.int2hex(int(amount*DECIMALS - vSize),'little',8))
    outUTXO = '00' + utils.int2hex(int(amount*DECIMALS - vSize),'little',8) + utils.encodeBech32AddressToHex(shareAddress) + utils.int2hex(rbf,'little',1) + utils.int2hex(locktime,'little',4)
    return res + outUTXO

def decodeCustomTx(hex:str):
    expectedOPReturn = hex[:2]
    
    hexLengthV1=hex[2:4]
    expectedDfTxMarkerV1 = hex[4:12]

    hexLengthV2=hex[4:6]
    expectedDfTxMarkerV2 = hex[6:14]
    if expectedOPReturn != '6a':
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Missing key word for OP_RETURN'
        }
    if expectedDfTxMarkerV1 == utils.stringToHex('DfTx'):
        hexLength = hexLengthV1
        txLock = None
        hex = hex[4:]  
    elif expectedDfTxMarkerV2 == utils.stringToHex('DfTx'):
        hexLength = hexLengthV2
        txLock = hex[2:4]
        hex = hex[6:] 
    else:
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Missing key word for DfTx marker'
        }

    if int(hexLength,16) != int(len(hex)/2):
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Indicated length in hex string does fit string length'
        }
    return CustomTxDecode.decodeCustomTx(hex[8:])

    


class TakeLoan(CustomTx.DfTx):
    def __init__(self,vaultId:str,address:str,amounts:list,rpcConnector):
        super().__init__()
        resAmounts = []
        for a in amounts:
            amount = a.split('@')
            resAmounts.append((float(amount[0]),utils.getTokenId(amount[1],rpcConnector)))
        self.params = [CustomTx.CustomTxType.TakeLoan,ParamType.Vault(vaultId),ParamType.Address(address),ParamType.Amounts(resAmounts)]

class AddPoolLiquidity(CustomTx.DfTx):
    def __init__(self,amountPerAddress:list[tuple[str,str]],shareAddress:str,rpcConnector):
        super().__init__()
        resAddressAmount = []
        for a in amountPerAddress:
            amount = a[1].split('@')
            resAddressAmount.append((a[0],float(amount[0]),utils.getTokenId(amount[1],rpcConnector)))
        self.params = [CustomTx.CustomTxType.AddPoolLiquidity,ParamType.AddressesWithAmount(resAddressAmount),ParamType.Address(shareAddress)]

@dataclass
class RPCConnector():
    NODE_URL:str
    NODE_USER:str
    NODE_PASSWORD:str


# if __name__ == '__main__':
#     #a = 'df1qamj6pp30h9y3syqcz8nwnekzg8dswul3tw5kjy'
#     #adr = utils.encodeBech32AddressToHex(a)

#     hex ='6a2c44665478706a0f400d0300000000001976a9148f83f4f005ba70f0c5d7a59d4696daee13a6e43b88ac010000'
#     #hex = '6a4c4f4466547869160014eee5a0862fb94918101811e6e9e6c241db0773f1000a00000000000000160014eee5a0862fb94918101811e6e9e6c241db0773f102ffffffffffffff7fffffffffffffff7f0105'
#     print(decodeCustomTx(hex))
    
    # print(hex(2059200))
    #print(utils.int2hex(144015528,'little',8))
    # #print(int(utils.convert_hex('b95f2f0300000000','little','big'),16)/100000000)
    # print(int(utils.convert_hex('90eac90100000000','big','little'),16))
    

    # # This is how this package could be used.
    # v = 'ed299afd57b7b354983efce9254436c7409e2c52a152d0a41bec0fb477f3a0b6'
    # a = 'tf1q6qj52ykxlf6halmx0g32gaumuuptactwgrqh23'

    # # a = 'df1q6qcutr37ex6xd5mjcmgrlr9239ck9z22605hfp'

    # RPCConn =RPCConnector("","","")
    # t = TakeLoan(v,a,["0.00112715@MSFT","1@DUSD"],RPCConn)
    # #print(t.hex())
    # p = AddPoolLiquidity([(a,"1@DUSD"),(a,"0.00112715@MSFT")],a,RPCConn)

    # input = {
    #     "txid":"b8639f6f27318c281f574ecb0360c1340e91e32780ca8da5c621a64a24026739",
    #     "vout":1,
    #     "amount":0.91485041
    # }
    # print(createRawTransaction([input],[t.hex()],a))
    # # print(t.hex())
    # # #d = DfTx(cTxType.TakeLoan,[v,a])
    # # #print(d.getHex())

    




import hashlib
#Defi base58 encoding and decoding.
#Based on https://bitcointalk.org/index.php?topic=1026.0 (public domain)

# for compatibility with following code...
class SHA256:
    new = hashlib.sha256

if str != bytes:
    # Python 3.x
    def ord(c):
        return c
    def chr(n):
        return bytes( (n,) )

__b58chars = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
__b58base = len(__b58chars)
b58chars = __b58chars

def b58encode(v):
    """ encode v, which is a string of bytes, to base58.
    """
    long_value = 0
    for (i, c) in enumerate(v[::-1]):
        if isinstance(c, str):
            c = ord(c)
        long_value += (256**i) * c

    result = ''
    while long_value >= __b58base:
        div, mod = divmod(long_value, __b58base)
        result = __b58chars[mod] + result
        long_value = div
    result = __b58chars[long_value] + result

    # Defi does a little leading-zero-compression:
    # leading 0-bytes in the input become leading-1s
    nPad = 0
    for c in v:
        if c == 0:
            nPad += 1
        else:
            break

    return (__b58chars[0]*nPad) + result

def b58decode(v, length = None):
    """ decode v into a string of len bytes
    """
    long_value = 0
    for i, c in enumerate(v[::-1]):
        pos = __b58chars.find(c)
        assert pos != -1
        long_value += pos * (__b58base**i)

    result = bytes()
    while long_value >= 256:
        div, mod = divmod(long_value, 256)
        result = chr(mod) + result
        long_value = div
    result = chr(long_value) + result

    nPad = 0
    for c in v:
        if c == __b58chars[0]:
            nPad += 1
            continue
        break

    result = bytes(nPad) + result
    if length is not None and len(result) != length:
        return None

    return result

def checksum(v):
    """Return 32-bit checksum based on SHA256"""
    return SHA256.new(SHA256.new(v).digest()).digest()[0:4]

def b58encode_chk(v):
    """b58encode a string, with 32-bit checksum"""
    return b58encode(v + checksum(v))

def b58decode_chk(v):
    """decode a base58 string, check and remove checksum"""
    result = b58decode(v)
    if result is None:
        return None
    if result[-4:] == checksum(result[:-4]):
        return result[:-4]
    else:
        return None

def get_bcaddress_version(strAddress):
    """ Returns None if strAddress is invalid.  Otherwise returns integer version of address. """
    addr = b58decode_chk(strAddress)
    if addr is None or len(addr)!=21:
        return None
    version = addr[0]
    return ord(version)


# if __name__ == '__main__':
#     t = '127742abf9be1e2a9e42d665faf9c982c65f413159'
     
#     t2 = bytes.fromhex(t)
#     # 
#     r = b58encode_chk(t2)
#     print(r)
#     print(get_bcaddress_version('8RxT1DhCdVUHb1DsYNWTxnGtevjqi1gUft'))
#     # a = 'dbTz4FC2N9YAQsgBo6u4o77GGT7oZktM6u'
#     r = b58decode('8RxT1DhCdVUHb1DsYNWTxnGtevjqi1gUft')
#     print(r.hex())
#     # dbTz4FC2N9YAQsgBo6u4o77GGT7oZktM6u
#     # dbTz4FC2N9YAQsgBo6u4o77GGT7oZktM6u
#     # 4VZ1cDYc4zw9PhE3tfpBKDaPiREpDpGN
#     # 1F7eKRrQsyKWi31enoV2gWFNuH8pCV8u8RQgK6v
#     # 1PrRqepz4usUe1XyxNje7r2MDTngRkN9mvcqGM


# import utils.utils as utils
# from utils.CustomTx import CustomTxType
import defichainUtils.utils.utils as utils
from defichainUtils.utils.CustomTx import CustomTxType
from dataclasses import asdict


def decodeCustomTx(hex):
    '''
    Hex String without OP_RETURN and DfTx Marker
    ''' 
    functionMarker = utils.hexToString(hex[:2])
    if functionMarker == 'j':
        res = decodeSetGovVariableHeight(hex[2:])
        if res != None:
            res['name'] = 'setgovheight'
    elif functionMarker == 's':
        res = decodePoolSwap(hex[2:])
        if res != None:
            res['name'] = 'poolswap'
    elif functionMarker == 'i':
        res = decodePoolSwapV2(hex[2:])
        if res != None:
            res['name'] = 'compositeswap'
    elif functionMarker == 'l':
        res = decodeAddPoolLiquidity(hex[2:])
        if res != None:
            res['name'] = 'addpoolliquidity'
    elif functionMarker == 'r':
        res = decodeRemovePoolLiquidity(hex[2:])
        if res != None:
            res['name'] = 'removepoolliquidity'
    elif functionMarker == 'p':
        res = decodeCreatePoolPair(hex[2:])
        if res != None:
            res['name'] = 'createpoolpair'
    else:
        c = CustomTxType()
        for key, value in asdict(c).items():
            if value == functionMarker:
                return {
                    'name': str(key).lower(),
                    'hex': hex[2:]
                }
        return None
    return res


def decodeSetGovVariableHeight(hex):
    # Only data part of hex string!
    govVariable = hex[2:2+2*int(hex[:2],16)]
    hex = hex[2+2*int(hex[:2],16):]
    if utils.stringToHex('ATTRIBUTES') == govVariable:
        # if hex[:2] != '01':
        #     return {
        #         'error': 1,
        #         'msg': 'Expected only one value in this function. Please contact package DEV.'
        #     }
        variables = [
            {'key':'attributes', 'type': [
                {'bytesCount': 4, 'key': None, 'type': int,'endian':None},
                {'bytesCount': 1, 'key': 'type', 'type': str,'endian':None},
                {'bytesCount': 1, 'key': 'typeId', 'type': int,'endian':None},
                {'bytesCount': 4, 'key': 'key', 'type': str,'endian':'big'},
                {'bytesCount': 4, 'key': 'keyId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'value', 'type': int,'endian':'little'}]
            },
            {'bytesCount': 4, 'key': 'height', 'type': int,'endian':'little'}
        ]
        return searchHex(variables,hex)
    elif utils.stringToHex('ICX_TAKERFEE_PER_BTC') == govVariable:
        # TODO: write Decoding
        pass
    elif utils.stringToHex('LP_LOAN_TOKEN_SPLITS') == govVariable:
        # TODO: write Decoding
        pass
    elif utils.stringToHex('LP_SPLITS') == govVariable:
        # TODO: write Decoding
        pass
    elif utils.stringToHex('ORACLE_DEVIATION') == govVariable:
        # TODO: write Decoding
        pass
    else:
        return None

def decodePoolSwap(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'fromTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'fromAmount', 'type': int,'endian':'big'},
        {'key': 'to', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'toTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'maxPriceCounter', 'type': int,'endian':'big'},
        {'bytesCount': 8, 'key': 'maxPriceDenominator', 'type': int,'endian':'big'}
    ]
    return searchHex(variables,hex)
def decodePoolSwapV2(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'fromTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'fromAmount', 'type': int,'endian':'big'},
        {'key': 'to', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'toTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'maxPriceCounter', 'type': int,'endian':'big'},
        {'bytesCount': 8, 'key': 'maxPriceDenominator', 'type': int,'endian':'big'},
        {'key':'pools', 'type':
            [{'bytesCount': 1, 'type': int,'endian':None}]
        }
    ]
    return searchHex(variables,hex)
def decodeAddPoolLiquidity(hex):
    variables = [
        {'key':'addresses', 'type': [
            {'key': 'from', 'type': 'address','endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'big'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'big'}
                ]}
        ]},
        {'key': 'shareAddress', 'type': 'address','endian':None}
    ]
    return searchHex(variables,hex)
def decodeRemovePoolLiquidity(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'tokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'big'}
    ]
    return searchHex(variables,hex)
def decodeCreatePoolPair(hex):
    variables = [
        {'bytesCount':1, 'key': 'tokenId', 'type': int,'endian':None},
        {'bytesCount':1, 'key': 'tokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'commission', 'type': int,'endian':'big'},
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'status', 'type': int,'endian':None}
    ]
    return searchHex(variables,hex)

def recursiveSearchHex(variable,hex):
    if type(variable['type']) == list:
        cnt = int(hex[:2],16)
        hex = hex[2:]
        arr = []
        for i in range(0,cnt):
            res = {}
            for idx,listVariable in enumerate(variable['type']):
                (key,value,hex) = recursiveSearchHex(listVariable,hex)
                if key is None and idx == len(variable['type']) - 1:
                    res = value
                else:
                    res[key] = value
            arr.append(res)
        return variable['key'],arr,hex
    else:
        return getValue(variable,hex)

def searchHex(variables:list,hex:str):
    res = {}
    for variable in variables:
        (key,value,hex) = recursiveSearchHex(variable,hex)
        res[key] = value
    
    if len(hex) == 0:
        res['hex'] = None
    else:
        res['hex'] = hex
    return res

def getValue(variable,hex):
    # Select subString from hex depending on length info in hex String or length info from variable input
    if variable['type'] == 'address':
        # START - Remove leading zeros
        while True:
            if len(hex) >= 2 and hex[:2] == '00':
                hex = hex[2:]
            else:
                break
        if len(hex) < 2:
            return None,None,None
        # END - Remove leading zeros
        hexLength = int(hex[:2],16)
        if hex[2:6] == '76a9' and hexLength == 25:
            addressType = 'base58'
            addressVersion = 18
            addressLength = int(hex[6:8],16)
            hexOfInterest = hex[8:8+2*addressLength]
            hex = hex[2+2*hexLength:]
        elif hex[2:4] == 'a9' and hexLength == 23: 
            addressType = 'base58'
            addressVersion = 90
            addressLength = int(hex[4:6],16)
            hexOfInterest = hex[6:6+2*addressLength]
            hex = hex[2+2*hexLength:]
        else:
            addressType = 'bech32'
            addressVersion = hex[2:4]
            addressLength = int(hex[4:6],16)
            hexOfInterest = hex[6:6+2*addressLength]
            hex = hex[2+2*hexLength:]
    else:
        hexOfInterest = hex[:2*variable['bytesCount']]
        hex = hex[2*variable['bytesCount']:]

    # Handle Endian Transformation
    if variable['endian'] == 'big':
        hexOfInterest = utils.convert_hex(hexOfInterest,'big','little')
    elif variable['endian'] == 'little':
        hexOfInterest = utils.convert_hex(hexOfInterest,'little','big')

    # Extract Value
    if variable['type'] == int:
        if hexOfInterest == '':
            valueOfInterest = 0
        else:
            valueOfInterest = int(hexOfInterest,16)
    elif variable['type'] == str:
        if hexOfInterest == '':
            valueOfInterest = hexOfInterest
        else:
            valueOfInterest = utils.hexToString(hexOfInterest)
    elif variable['type'] == 'address':
        valueOfInterest = utils.decodeHexToAddress(hexOfInterest,addressType,addressVersion)
    else:
        valueOfInterest = None
    
    if 'key' not in variable:
        return None,valueOfInterest,hex
    if variable['key'] != None:
        return variable['key'],valueOfInterest,hex
    return None,None,hex


from __future__ import annotations
from dataclasses import dataclass

@dataclass
class AttributeType:
    Live      = 'l'
    Oracles   = 'o'
    Param     = 'a'
    Token     = 't'
    Poolpairs = 'p'
    Locks     = 'L'

@dataclass
class ParamIDs:
    DFIP2201  = 'a'
    DFIP2203  = 'b'
    TokenID   = 'c'
    Economy   = 'e'
    DFIP2206A = 'f'
    DFIP2206F = 'g'

@dataclass
class OracleIDs:
    Splits    = 'a'

@dataclass
class EconomyKeys:
    PaybackDFITokens  = 'a'
    PaybackTokens     = 'b'
    DFIP2203Current   = 'c'
    DFIP2203Burned    = 'd'
    DFIP2203Minted    = 'e'
    DFIP2206FCurrent  = 'f'
    DFIP2206FBurned   = 'g'
    DFIP2206FMinted   = 'h'
    DexTokens         = 'i'

@dataclass
class DFIPKeys:
    Active           = 'a'
    Premium          = 'b'
    MinSwap          = 'c'
    RewardPct        = 'd'
    BlockPeriod      = 'e'
    DUSDInterestBurn = 'g'
    DUSDLoanBurn     = 'h'
    StartBlock       = 'i'

@dataclass
class TokenKeys:
    PaybackDFI            = 'a'
    PaybackDFIFeePCT      = 'b'
    LoanPayback           = 'c'
    LoanPaybackFeePCT     = 'd'
    DexInFeePct           = 'e'
    DexOutFeePct          = 'f'
    DFIP2203Enabled       = 'g'
    FixedIntervalPriceId  = 'h'
    LoanCollateralEnabled = 'i'
    LoanCollateralFactor  = 'j'
    LoanMintingEnabled    = 'k'
    LoanMintingInterest   = 'l'
    Ascendant             = 'm'
    Descendant            = 'n'
    Epitaph               = 'o'

@dataclass
class PoolKeys:
    TokenAFeePCT = 'a'
    TokenBFeePCT = 'b'
    TokenAFeeDir = 'c'
    TokenBFeeDir = 'd'



from __future__ import annotations
from dataclasses import dataclass

@dataclass
class Vault:
    value:str

@dataclass
class Address:
    value:str

# @dataclass
# class Addresses:
#     value:list[str]

@dataclass
class AddressesWithAmount:
    '''
    List of Adress Amount Tuples. Tuple has str (address), float (number of dTokens) and int (ID of dToken)
    '''
    value:list[tuple[str,float,int]]
    
    def toDict(self):
        d = {}
        for v in self.value:
            if v[0] not in d:
                d[v[0]] = {}

            if v[2] not in d[v[0]]:
                d[v[0]][v[2]] = v[1]
            else:
                d[v[0]][v[2]] = d[v[0]][v[2]] + v[1]
        return d

@dataclass
class Amounts:
    '''
    List of Amount Tuples. Amount Tuple has float (number of dTokens) and int (ID of dToken)
    '''
    value:list[tuple[float, int]]